﻿
using Telerik.Windows.Controls.Map;
using System;
using System.Collections.Generic;
using Microsoft.SqlServer.Types;
using System.Windows;
using System.Windows.Media;
using System.Linq;
using System.Windows.Controls;

namespace OA.Spatial.Example1.WPF
{
    public static class SqlSpatialHelper
    {
        // <summary>
        /// Converts an object into a RadMap Object based on the input SqlGeography type
        /// </summary>
        /// <typeparam name="TEntity">Entity type containing sql geometry</typeparam>
        /// <param name="entities">the collection of entites to plot</param>
        /// <param name="geometryProperty">Which property contains SqlGeometry</param>
        /// <returns>List of the created map objects</returns>
        public static IList<FrameworkElement> LoadSpatial<TEntity>(this Telerik.Windows.Controls.Map.InformationLayer infoLayer,
                                                                  IList<TEntity> entities,
                                                                  Func<TEntity, SqlGeography> geometryProperty)
            where TEntity : class
        {
            return infoLayer.LoadSpatial(entities, geometryProperty, null);
        }


        // <summary>
        /// Converts an object into a RadMap Object based on the input SqlGeography type
        /// </summary>
        /// <typeparam name="TEntity">Entity type containing sql geometry</typeparam>
        /// <param name="entities">the collection of entites to plot</param>
        /// <param name="geometryProperty">Which property contains SqlGeometry</param>
        /// <param name="extendedData">Create the extended data for the map shape</param>
        /// <returns>List of the created map objects</returns>
        public static IList<FrameworkElement> LoadSpatial<TEntity>(this Telerik.Windows.Controls.Map.InformationLayer infoLayer,
                                                                    IList<TEntity> entities,
                                                                    Func<TEntity, SqlGeography> geometryProperty,
                                                                    Func<TEntity, ExtendedData> extendedData)
            where TEntity : class
        {
            var query = entities.Select(sc => new SqlShapeDTO
            {
                Geometry = geometryProperty(sc),
                ExtendedData = extendedData == null ? null : extendedData(sc)
            }).ToList();

            var shapeElements = query.ToMapElements();

            shapeElements.ToList()
                         .ForEach( element => infoLayer.Items.Add(element));

            return shapeElements;
        }

        public static void SetToolTip<T>(this FrameworkElement element, Func<T, string> tooltipBuilder) where T : class
        {
            var extendedData = element.Tag as T;

            if (extendedData != null)
            {
                ToolTip toolTip = new ToolTip();
                toolTip.Content = tooltipBuilder(extendedData);

                element.ToolTip = toolTip;
            }
        }

        private static IList<FrameworkElement> ToMapElements(this IEnumerable<SqlShapeDTO> records)
        {
            IList<FrameworkElement> elements = new List<FrameworkElement>();

            foreach (var record in records)
            {
                if (record == null)
                    continue;

                OpenGisGeometryType geometryType =
                (OpenGisGeometryType)Enum.Parse(typeof(OpenGisGeometryType), (string)record.Geometry.STGeometryType());

                switch (geometryType)
                {
                    case OpenGisGeometryType.MultiPoint:
                        AddPoints(record, elements);
                        break;
                    case OpenGisGeometryType.Point:
                        AddPoint(record, elements);
                        break;
                    case OpenGisGeometryType.MultiPolygon:
                    case OpenGisGeometryType.Polygon:
                        AddPolygon(record, elements);
                        break;
                    default:
                        break;
                }
            }

            return elements;
        }

        private static void AddPoints(SqlShapeDTO record, IList<FrameworkElement> mapObjects)
        {
            for (int i = 0; i < record.Geometry.STNumGeometries(); i++)
            {
                AddPoint(record, mapObjects);
            }
        }

        private static void AddPoint(SqlShapeDTO record, IList<FrameworkElement> mapObjects)
        {
            double longitude = (double)record.Geometry.Long;
            double latitude = (double)record.Geometry.Lat;

            var point = CreatePoint(record, latitude, longitude);

            mapObjects.Add(point);
        }

        private static FrameworkElement CreatePoint(SqlShapeDTO record, double latitude, double longitude)
        {
            var point = new MapPinPoint();
            point.ExtendedData = record.ExtendedData;
            MapLayer.SetLocation(point, new Location(latitude, longitude));
            return point;
        }

        private static void CreateMapPath(SqlShapeDTO record, IList<LocationCollection> locations, IList<FrameworkElement> mapObjects)
        {
            MapPathGeometry geometry = CreateMapPathGeometry(locations[0]);
            locations.RemoveAt(0);
            foreach (LocationCollection subLocations in locations)
            {
                AddFigure(geometry, subLocations);
            }

            MapPath path = new MapPath();
            path.ExtendedData = record.ExtendedData;

            path.Data = geometry;

            mapObjects.Add(path);
        }

        private static void AddFigure(MapPathGeometry geometry, LocationCollection locations)
        {
            MapPolyLineSegment segment = new MapPolyLineSegment();
            segment.Points = locations;

            MapPathFigure figure = new MapPathFigure();
            if (segment.Points != null && segment.Points.Count > 0)
            {
                figure.StartPoint = segment.Points[0];
            }
            figure.Segments.Add(segment);
            geometry.Figures.Add(figure);

            // add empty figure to correct odd position for each inner polygon
            figure = new MapPathFigure();
            if (segment.Points != null && segment.Points.Count > 0)
            {
                figure.StartPoint = segment.Points[segment.Points.Count - 1];
            }
            geometry.Figures.Add(figure);
        }

        private static MapPathGeometry CreateMapPathGeometry(LocationCollection locations)
        {
            MapPolyLineSegment polyline = new MapPolyLineSegment();
            polyline.Points = locations;


            MapPathFigure figure = new MapPathFigure();
            figure.StartPoint = locations[0];
            figure.Segments.Add(polyline);

            MapPathGeometry geometry = new MapPathGeometry();
            geometry.FillRule = FillRule.EvenOdd;
            geometry.Figures.Add(figure);

            return geometry;
        }

        private static void AddPolygon(SqlShapeDTO record, IList<FrameworkElement> mapObjects)
        {
            IList<LocationCollection> points = GetPoints(record.Geometry);
            if (points.Count < 2)
            {
                var polygon = new MapPolygon();

                polygon.Points = points[0];

                polygon.ExtendedData = record.ExtendedData;

                mapObjects.Add(polygon);
            }
            else
            {
                CreateMapPath(record, points, mapObjects);
            }
        }


        private static IList<LocationCollection> GetPoints(SqlGeography record)
        {
            List<LocationCollection> pointsCollection = new List<LocationCollection>();
            for (int i = 1; i <= (int)record.STNumGeometries(); i++)
            {
                LocationCollection points = new LocationCollection();

                var ptGeometry = record.STGeometryN(i);

                // Add line segments 
                for (int j = 1; j <= ptGeometry.STNumPoints(); j++)
                {
                    var point = ptGeometry.STPointN(j);
                    points.Add(new Location((double)point.Lat, (double)point.Long));
                }

                pointsCollection.Add(points);
            }
            return pointsCollection;

        }

        /// <summary>
        /// Temporary object used to create map shapes
        /// </summary>
        private class SqlShapeDTO
        {
            public SqlGeography Geometry
            {
                get;
                set;
            }

            public ExtendedData ExtendedData
            {
                get;
                set;
            }
        }
    }
}